@@ -1,7 +1,7 @@
 /* SPDX-License-Identifer: GPL-2.0-or-later
 
 Copyright (C) 2014  Vyacheslav Trushkin
-Copyright (C) 2020,2021  Boian Bonev
+Copyright (C) 2020-2022  Boian Bonev
 
 This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 
@@ -82,6 +82,8 @@ static inline void view_batch(struct xxxid_stats_arr *cs,struct xxxid_stats_arr
 }
 
 inline void view_batch_init(void) {
+	if (!read_task_delayacct())
+		printf("Warning: task_delayacct is 0, enable by: echo 1 > /proc/sys/kernel/task_delayacct\n");
 }
 
 inline void view_batch_fini(void) {
